const Applet = imports.ui.applet;
const Mainloop = imports.mainloop;
const Lang = imports.lang;
const Cinnamon = imports.gi.Cinnamon;
const GnomeSession = imports.misc.gnomeSession;
const St = imports.gi.St;
const Clutter = imports.gi.Clutter;
const Main = imports.ui.main;
const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;

const PopupMenu = imports.ui.popupMenu;
const AppFavorites = imports.ui.appFavorites;
const Gtk = imports.gi.Gtk;
const ScreenSaver = imports.misc.screenSaver;
const Util = imports.misc.util;
const Tweener = imports.ui.tweener;
const DND = imports.ui.dnd;
const Meta = imports.gi.Meta;
const AccountsService = imports.gi.AccountsService;
const Gettext = imports.gettext;
const FileUtils = imports.misc.fileUtils;
const Signals = imports.signals;

const Settings = imports.ui.settings;
var APIMenu;
try {
   APIMenu = imports.gi.CMenu;
} catch(e) {
   APIMenu = imports.gi.GMenu;
}
const CMenu = APIMenu;
const AppletDir = imports.ui.appletManager.applets['menu@cdos.org'];
if(typeof PopupMenu.PopupIconMenuItem != 'function') {
    let iconMenu = AppletDir.popupiconmenu;
    PopupMenu.PopupIconMenuItem = iconMenu.PopupIconMenuItem;
}

let appsys = Cinnamon.AppSystem.get_default();

const ICON_SIZE = 16; 
const MAX_FAV_ICON_SIZE = 32;
const CATEGORY_ICON_SIZE = 32;
const APPLICATION_ICON_SIZE = 16;
const HOVER_ICON_SIZE =48;

const USER_DESKTOP_PATH = FileUtils.getUserDesktopDir();

var Session = new GnomeSession.SessionManager();

function _(str) {
   let resultConf = Gettext.dgettext("menu@cdos.org", str);
   if(resultConf != str) {
      return resultConf;
   }
   return Gettext.gettext(str);
}

function ApplicationContextMenuItem(appButton, label, action) {
    this._init(appButton, label, action);
}

ApplicationContextMenuItem.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function (appButton, label, action) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            focusOnHover: false
        });

        this._appButton = appButton;
        this._action = action;
        this.label = new St.Label({
            text: label
        });
        this.addActor(this.label);
    },

    activate: function (event) {
        this._appButton.toggleMenu();
        switch (this._action) {
        case "add_to_panel":
            let winListApplet = false;
            try{
                winListApplet = imports.ui.appletManager.applets['window-list@cdos.org'];
            }catch (e) {
            }
            if (winListApplet)
                winListApplet.applet.GetAppFavorites().addFavorite(this._appButton.app.get_id());
            else{
                let settings = new Gio.Settings({
                    schema: 'org.cdos'
                });
                let desktopFiles = settings.get_strv('panel-launchers');
                desktopFiles.push(this._appButton.app.get_id());
                settings.set_strv('panel-launchers', desktopFiles);
            }
            break;
        case "add_to_desktop":
            let file = Gio.file_new_for_path(this._appButton.app.get_app_info().get_filename());
            let destFile = Gio.file_new_for_path(USER_DESKTOP_PATH + "/" + this._appButton.app.get_id());
            try {
                file.copy(destFile, 0, null, function () {});
                // Need to find a way to do that using the Gio library, but modifying the access::can-execute attribute on the file object seems unsupported
                Util.spawnCommandLine("chmod +x \"" + USER_DESKTOP_PATH + "/" + this._appButton.app.get_id() + "\"");
            } catch (e) {
                global.log(e);
            }
            break;
        case "add_to_favorites":
            AppFavorites.getAppFavorites().addFavorite(this._appButton.app.get_id());
            break;
        case "remove_from_favorites":
            AppFavorites.getAppFavorites().removeFavorite(this._appButton.app.get_id());
            break;
        }
        return false;
    }
};

function GenericApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

GenericApplicationButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (appsMenuButton, app, withMenu) {
        this.app = app;
        this.appsMenuButton = appsMenuButton;
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });

        this.withMenu = withMenu;

        if (this.withMenu) {
            this.menu = new ApplicationMenu(this.actor);
            this.menu.actor.set_style_class_name('menu-context-menu');
        }
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
        if (event.get_button() == 2) {
            this.appsMenuButton._clearSelections(this.appsMenuButton.categoriesBox);
        }
        if (event.get_button() == 3) {
            this.appsMenuButton.closeApplicationsContextMenus(this.app, true);
            this.toggleMenu();
        }
        return true;
    },

    activate: function (event) {
        this.app.open_new_window(-1);
        this.appsMenuButton.menu.close();
    },

    closeMenu: function () {
        if (this.withMenu) this.menu.close();
    },

    toggleMenu: function () {
        if (!this.withMenu) return;

        if (!this.menu.isOpen) {
            let children = this.menu.box.get_children();
            for (var i in children) {
                children[i].destroy();
            }
            let menuItem;
            menuItem = new ApplicationContextMenuItem(this, _("Add to panel"), "add_to_panel");
            this.menu.addMenuItem(menuItem);
            if (USER_DESKTOP_PATH) {
                menuItem = new ApplicationContextMenuItem(this, _("Add to desktop"), "add_to_desktop");
                this.menu.addMenuItem(menuItem);
            }
            if (AppFavorites.getAppFavorites().isFavorite(this.app.get_id())) {
                menuItem = new ApplicationContextMenuItem(this, _("Remove from favorites"), "remove_from_favorites");
                this.menu.addMenuItem(menuItem);
            } else {
                menuItem = new ApplicationContextMenuItem(this, _("Add to favorites"), "add_to_favorites");
                this.menu.addMenuItem(menuItem);
            }
        }
        this.menu.toggle();
    },
}

function ApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

ApplicationButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function (appsMenuButton, app) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);

        this.actor.add_style_class_name('menu-category-button');
        this.icon = this.app.create_icon_texture(APPLICATION_ICON_SIZE);
        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.icon);
        this.addActor(this.label);

        this._draggable = DND.makeDraggable(this.actor);
        this.isDraggableApp = true;
    },

    get_app_id: function () {
        return this.app.get_id();
    },

    getDragActor: function () {
        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let nbFavorites = favorites.length;
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let real_size = (0.7 * monitorHeight) / nbFavorites;
        let icon_size = 0.6 * real_size;
        if (icon_size > MAX_FAV_ICON_SIZE) icon_size = MAX_FAV_ICON_SIZE;
        return this.app.create_icon_texture(icon_size);
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function () {
        return this.actor;
    }
};
Signals.addSignalMethods(ApplicationButton.prototype);

function PlaceButton(appsMenuButton, place, button_name) {
    this._init(appsMenuButton, place, button_name);
}

PlaceButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function (appsMenuButton, place, button_name) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.place = place;
        this.button_name = button_name;
        this.label = new St.Label({
            text: this.button_name,
            style_class: 'menu-application-button-label'
        });
        this.icon = place.iconFactory(APPLICATION_ICON_SIZE);
        this.addActor(this.icon);
        this.addActor(this.label);
    },

    activate: function () {
        this.place.launch();
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceButton.prototype);

function ApplicationMenu() {
    this._init.apply(this, arguments);
}

ApplicationMenu.prototype = {
    __proto__: PopupMenu.PopupMenuBase.prototype,

    _init: function (sourceActor) {
        PopupMenu.PopupMenuBase.prototype._init.call(this, sourceActor);

        this.actor = new St.BoxLayout({ style_class: 'popup-sub-menu'});

        this.actor.add_actor(this.box);
        this.actor._delegate = this;
        this.actor.clip_to_allocation = true;
        //this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
        this.actor.hide();
    },

    open: function(animate) {
        if (this.isOpen)
            return;

        this.isOpen = true;

        this.actor.show();

        if (animate) {
            let [minHeight, naturalHeight] = this.actor.get_preferred_height(-1);
            this.actor.height = 0;
            Tweener.addTween(this.actor,
                             { height: naturalHeight,
                               time: 0.25,
                               onCompleteScope: this,
                               onComplete: function() {
                                   this.actor.set_height(-1);
                                   this.emit('open-state-changed', true);
                               }
                             });
        } else {
            this.emit('open-state-changed', true);
        }
    },

    close: function(animate) {
        if (!this.isOpen)
            return;

        this.isOpen = false;


        if (animate) {
            Tweener.addTween(this.actor,
                             { height: 0,
                               time: 0.25,
                               onCompleteScope: this,
                               onComplete: function() {
                                   this.actor.hide();
                                   this.actor.set_height(-1);
                                   this.emit('open-state-changed', false);
                               },
                             });
            } else {
                this.actor.hide();

                this.isOpen = false;
                this.emit('open-state-changed', false);
            }
    },

};

function CategoryButton(category) {
    this._init(category);
}

CategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (category) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        var label;
        this.actor.add_style_class_name('menu-category-button');
        if (category) {
            let icon = category.get_icon();
            if (icon && icon.get_names) this.icon_name = icon.get_names().toString();
            else this.icon_name = "";
            label = category.get_name();
        } else label = _("All Applications");
        this.label = new St.Label({
            text: label,
            style_class: 'menu-category-button-label'
        });
        if (category && this.icon_name) {
            this.icon = new St.Icon({
                icon_name: this.icon_name,
                icon_size: CATEGORY_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            this.addActor(this.icon);
        }
        this.addActor(this.label);
        //this.actor.set_tooltip_text(category.get_name());
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.set_style_class_name('menu-context-menu');
    },

    activate: function () {
        this.menu.toggle();
    },

    removeItems: function () {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(CategoryButton.prototype);

function PlaceCategoryButton() {
    this._init();
}

PlaceCategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function () {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.label = new St.Label({
            text: _("Places"),
            style_class: 'menu-category-button-label'
        });
        this.icon = new St.Icon({
            icon_name: "folder",
            icon_size: CATEGORY_ICON_SIZE,
            icon_type: St.IconType.FULLCOLOR
        });
        this.addActor(this.icon);
        this.addActor(this.label);
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.set_style_class_name('menu-context-menu');
    },

    activate: function () {
        this.menu.toggle();
    },

    removeItems: function () {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceCategoryButton.prototype);

function FavoritesButton(appsMenuButton, app, nbFavorites) {
    this._init(appsMenuButton, app, nbFavorites);
}

FavoritesButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function (appsMenuButton, app, nbFavorites) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let real_size = (0.7 * monitorHeight) / nbFavorites;
        let icon_size = 0.6 * real_size;
        if (icon_size > MAX_FAV_ICON_SIZE) icon_size = MAX_FAV_ICON_SIZE;
        this.actor.style = "padding-top: " + (icon_size / 3) + "px;padding-bottom: " + (icon_size / 3) + "px; margin:auto;"

        this.actor.add_style_class_name('menu-favorites-button');
        this.addActor(app.create_icon_texture(icon_size));

        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.label);

        this._draggable = DND.makeDraggable(this.actor);
        this.isDraggableApp = true;
    },

    get_app_id: function () {
        return this.app.get_id();
    },

    getDragActor: function () {
        return new Clutter.Clone({
            source: this.actor
        });
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function () {
        return this.actor;
    }
};

function TextBoxItem(label, icon, func, parent, hoverIcon) {
    this._init(label, icon, func, parent, hoverIcon);
}

TextBoxItem.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (label, icon_name, func, parent, hoverIcon) {
        this.parent = parent;
        this.hoverIcon = hoverIcon;
        this.icon_name = icon_name;
    /*  //change
	    this.icon = new St.Icon({
            style_class: 'popup-menu-icon',
            icon_type: St.IconType.FULLCOLOR,
            icon_name: icon_name,
            icon_size: ICON_SIZE
        });
        */
        this.func = func;
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, null);

        this.actor.add_style_class_name('menu-category-button');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);
        this.label = new St.Label({
            text: label,
            style_class: 'menu-application-button-label'
        });
        //this.addActor(this.icon);  //change
        this.addActor(this.label);
    },

    setActive: function (active) {
        if (active) {
            this.actor.remove_style_class_name('menu-category-button');
            this.actor.add_style_class_name('menu-category-button-selected');
            this.hoverIcon._refresh(this.icon_name);
        }
        else {
            this.actor.remove_style_class_name('menu-category-button-selected');            
            this.actor.add_style_class_name('menu-category-button');
            this.hoverIcon._refresh();
        }
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
    },

    activate: function (event) {
        eval(this.func);
        this.parent.close();
    }
};

function AllProgramsItem(label, icon, parent, rightP, leftP) {
    this._init(label, icon, parent, rightP, leftP);
}

AllProgramsItem.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (label, icon, parent) {
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, label);

        this.actor.set_style_class_name('menu-category-button');
        this.parent = parent;
        //this.actor.set_style_class_name('');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);
	    //this.label = new St.Label({text: label});
        this.label.realize();
	    //this.icon = new St.Icon({style_class: 'popup-menu-icon', icon_type: St.IconType.FULLCOLOR, icon_name: icon, icon_size: ICON_SIZE });
	    //this.addActor(this.icon); mm
	    this.addActor(this.label);
    },

    setActive: function(active){
        if (active)
            this.actor.add_style_class_name('menu-category-button-selected');
        else
            this.actor.remove_style_class_name('menu-category-button-selected');
    },

    _onButtonReleaseEvent: function(actor, event){
        if (event.get_button()==1){
            this.activate(event);
        }
    },

    activate: function(event){
        if (this.parent.leftPane.get_child() == this.parent.favsBox) {
            this.parent.leftPane.set_child(this.parent.categoriesScrollBox, {
                y_align: St.Align.END,
                y_fill: false
            });
            this.label.set_text(" Favorites");
            this.parent.selectedAppBox.show();
            this.parent._appletStyles();
        }else {
            this.parent.leftPane.set_child(this.parent.favsBox, {
                y_align: St.Align.END,
                y_fill: false
            });
            this.label.set_text(" All Programs");
            this.parent.selectedAppBox.hide();
        }
    }
};

function HoverIcon(menu) {
    this._init(menu);
}

HoverIcon.prototype = {
    _init: function (menu) {
        try {
            this._menu = menu;
            let userBox = new St.BoxLayout({ style_class: 'user-box', reactive: true, vertical: false });
            this._userIcon = new St.Bin({ style_class: 'user-icon'});
            userBox.connect('button-press-event', Lang.bind(this, function() {
                this._menu.toggle();
                Util.spawnCommandLine("cdos-settings user");
            }));

            this._userIcon.hide();
            userBox.add(this._userIcon,
                      { x_fill:  true,
                        y_fill:  false,
                        x_align: St.Align.END,
                        y_align: St.Align.START });
            this.userLabel = new St.Label(({ style_class: 'user-label'}));
            userBox.add(this.userLabel,
                    { x_fill:  true,
                      y_fill:  false,
                      x_align: St.Align.END,
                      y_align: St.Align.MIDDLE });
            this._user = AccountsService.UserManager.get_default().get_user(GLib.get_user_name());
            this._userLoadedId = this._user.connect('notify::is_loaded', Lang.bind(this, this._onUserChanged));
            this._userChangedId = this._user.connect('changed', Lang.bind(this, this._onUserChanged));
            this.actor = new St.Bin({style_class: 'hover-icon'});
            this.userBox = userBox;
            this._icon = new St.Icon({
                icon_size: HOVER_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            this.actor.show();
            this.actor.set_child(this.userBox);
            this._onUserChanged();
        }
        catch(e) {
            global.logError(e);
        }

    },
  
    _onUserChanged: function() {
        this._icon.hide();
        this.userBox.show();
        if (this._user.is_loaded) {
            this.userLabel.set_text (GLib.get_user_name());
            if (this._userIcon) {
                let iconFileName = this._user.get_icon_file();
                let iconFile = Gio.file_new_for_path(iconFileName);
                let icon;
                if (iconFile.query_exists(null)) {
                    icon = new Gio.FileIcon({file: iconFile});
                } else {
                    icon = new Gio.ThemedIcon({name: 'avatar-default'});
                }
                let img = St.TextureCache.get_default().load_gicon(null, icon, HOVER_ICON_SIZE);
                this._userIcon.set_child (img);
                this.actor.set_child(this.userBox);
                this._userIcon.show();
            }
        }
    },
    _refresh: function (icon_name) {
        if(icon_name) {
            this.userBox.hide();
            this.actor.set_child(this._icon);
            this._icon.set_icon_name(icon_name);
            this._icon.show();
        }
        else {
            this.actor.set_child(this.userBox);
            this.userBox.show();
            //this._userIcon.show();
            this._icon.hide();
        }
    }
};
/*
function ShutdownContextMenuItem(parentMenu, menu, label, action) {
    this._init(parentMenu, menu, label, action);
}

ShutdownContextMenuItem.prototype = {
    __proto__: ApplicationContextMenuItem.prototype,

    _init: function (parentMenu, menu, label, action) {
        this.parentMenu = parentMenu;
        ApplicationContextMenuItem.prototype._init.call(this, menu, label, action);
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
    },

    activate: function (event) {
        switch (this._action) {
        case "logout":
            this._session.LogoutRemote(0);
            break;
        case "lock":
            this._screenSaverProxy.LockRemote();
            break;
        }
        this._appButton.toggle();
        this.parentMenu.toggle();
        return false;
    }

};
*/
function ShutdownContextMenu(position) {
    this._init(position);
}

ShutdownContextMenu.prototype = {

    _init: function (position) {
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
        
        this.actor = new St.Bin({ style_class: 'shutdown-context-menu'});
        this.actor.set_position(position[0], position[1]);

        //let menu = new ApplicationMenu(this.actor);
        //menu.actor.remove_style_class_name("popup-sub-menu");
        let menu = new St.BoxLayout({ vertical: true});
        let item;
        item = new PopupMenu.PopupIconMenuItem(_("Lock Screen"), "lock-screen", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            let screensaver_settings = new Gio.Settings({schema: "org.cdos.desktop.screensaver"});
            let screensaver_dialog   = Gio.file_new_for_path("/usr/bin/cdos-screensaver-command");
            if(screensaver_dialog.query_exists(null)) {
                if(screensaver_settings.get_boolean("ask-for-away-message")) {
                    Util.spawnCommandLine("cdos-screensaver-lock-dialog");
                }
                else {
                    Util.spawnCommandLine("cdos-screensaver-command --lock");
                }
            }
        }));
        menu.add_actor(item.actor);
        if (GLib.getenv("XDG_SEAT_PATH")) {
            // LightDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-greeter");
            }));
            menu.addMenuItem(item);

            item = new PopupMenu.PopupIconMenuItem(_("Guest Session"), "guest-session", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-guest");
            }));
            menu.add_actor(item.actor);
            //menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/mdmflexiserver", GLib.FileTest.EXISTS)) {
            // MDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("mdmflexiserver");
            }));
            menu.add_actor(item.actor);
            
            //menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/gdmflexiserver", GLib.FileTest.EXISTS)) {
            // GDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("gdmflexiserver");
                }));
            menu.add_actor(item.actor);
            //menu.addMenuItem(item);
        }
        item = new PopupMenu.PopupIconMenuItem(_("Log Out..."), "logout", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.LogoutRemote(0);
        }));
        //menu.addMenuItem(item);
        menu.add_actor(item.actor);

        //menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

        item = new PopupMenu.PopupIconMenuItem(_("Power Off..."), "shutdown", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.ShutdownRemote();
        }));
        menu.add_actor(item.actor);
        //menu.addMenuItem(item);
        //menu.toggle();
        this.actor.add_actor(menu);
    }
};
function ShutdownMenu(parent, hoverIcon) {      /*关机菜单*/
    this._init(parent, hoverIcon);
}

ShutdownMenu.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (parent, hoverIcon) {
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
        let label = '';
        this.hoverIcon = hoverIcon;
        this.parent = parent;
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, label);
        this.actor.set_style_class_name('menu-category-button');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);

	    this.icon = new St.Icon({style_class: 'popup-menu-icon',
                                 icon_type: St.IconType.FULLCOLOR,
                                 icon_name: 'forward',
                                 icon_size: ICON_SIZE });
	    this.addActor(this.icon);
       
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.remove_style_class_name("popup-sub-menu");
        /*
        let menuItem;
        menuItem = new ShutdownContextMenuItem(this.parent, this.menu, _("Logout"), "logout");
        this.menu.addMenuItem(menuItem);
        menuItem = new ShutdownContextMenuItem(this.parent, this.menu, _("Lock Screen"), "lock");
        this.menu.addMenuItem(menuItem);
        */
        let item;
        item = new PopupMenu.PopupIconMenuItem(_("Lock Screen"), "lock-screen", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            let screensaver_settings = new Gio.Settings({schema: "org.cdos.desktop.screensaver"});
            let screensaver_dialog   = Gio.file_new_for_path("/usr/bin/cdos-screensaver-command");
            if(screensaver_dialog.query_exists(null)) {
                if(screensaver_settings.get_boolean("ask-for-away-message")) {
                    Util.spawnCommandLine("cdos-screensaver-lock-dialog");
                }
                else {
                    Util.spawnCommandLine("cdos-screensaver-command --lock");
                }
            }
        }));
        this.menu.addMenuItem(item);
        if (GLib.getenv("XDG_SEAT_PATH")) {
            // LightDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-greeter");
            }));
            this.menu.addMenuItem(item);

            item = new PopupMenu.PopupIconMenuItem(_("Guest Session"), "guest-session", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-guest");
            }));
            this.menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/mdmflexiserver", GLib.FileTest.EXISTS)) {
            // MDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("mdmflexiserver");
            }));
            this.menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/gdmflexiserver", GLib.FileTest.EXISTS)) {
            // GDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("gdmflexiserver");
                }));
                this.menu.addMenuItem(item);
        }
        item = new PopupMenu.PopupIconMenuItem(_("Log Out..."), "logout", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.LogoutRemote(0);
        }));
        this.menu.addMenuItem(item);

        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

        item = new PopupMenu.PopupIconMenuItem(_("Power Off..."), "shutdown", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.ShutdownRemote();
        }));
        this.menu.addMenuItem(item);
    },

    setActive: function (active) {
        if (active) { 
            this.actor.set_style_class_name('menu-category-button-selected');
            this.hoverIcon._refresh('system-log-out');
        }
        else this.actor.set_style_class_name('menu-category-button');
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            let pos = event.get_coords();
            /*
            this._menuManager = new PopupMenu.PopupMenuManager(this);
            this._menu = new Applet.AppletPopupMenu(this, 3);
            this._menuManager.addMenu(this._menu);
            let section = new PopupMenu.PopupMenuSection();
            this._menu.addMenuItem(section);
            let pmenu = new ShutdownContextMenu(pos);
            section.actor.add_actor(pmenu.actor);
            */
            this.menu.toggle();
            //let pmenu = new ShutdownContextMenu(pos);
            //Main.notify(pos.toString() + pos[0]);

            //this._menu.toggle();
        }

    }
};



function RightButtonsBox(appsMenuButton, menu) {   /*右边按钮：主目录～Shutdown*/
    this._init(appsMenuButton, menu);
}

RightButtonsBox.prototype = {
    _init: function (appsMenuButton, menu) {
        this.appsMenuButton = appsMenuButton;
        this.actor = new St.BoxLayout({
            style_class: 'right-buttons-box',
            vertical: true
        });
        this.shutdownBox = new St.BoxLayout({
            vertical: false
        });
        this.actor._delegate = this;
        this.menu = menu;
        this.addItems();
    },

    addItems: function () {
        this.hoverIcon = new HoverIcon(this.menu);
        this.home = new TextBoxItem(_("Home"), "folder-home", "Util.spawnCommandLine('nemo')", this.menu, this.hoverIcon, false);
        this.documents = new TextBoxItem(_("Documents"), "folder-documents", "Util.spawnCommandLine('nemo Documents')", this.menu, this.hoverIcon, false);
        this.pictures = new TextBoxItem(_("Pictures"), "folder-pictures", "Util.spawnCommandLine('nemo Pictures')", this.menu, this.hoverIcon, false);
        this.music = new TextBoxItem(_("Music"), "folder-music", "Util.spawnCommandLine('nemo Music')", this.menu, this.hoverIcon, false);
        this.videos = new TextBoxItem(_("Videos"), "folder-videos", "Util.spawnCommandLine('nemo Videos')", this.menu, this.hoverIcon, false);
        this.computer = new TextBoxItem(_("Computer"), "computer", "Util.spawnCommandLine('nemo computer:///')", this.menu, this.hoverIcon, false);
        //this.packageItem = new Items.TextBoxItem(_("Package Manager"), "synaptic", "Util.spawnCommandLine('gksu synaptic')", this.menu, this.hoverIcon, false);
        this.control = new TextBoxItem(_("Control Center"), "gnome-control-center", "Util.spawnCommandLine('cdos-control-center')", this.menu, this.hoverIcon, false);
        this.run = new TextBoxItem(_("Run"), "system-run", "Main.getRunDialog().open()", this.menu, this.hoverIcon, false);
        this.terminal = new TextBoxItem(_("Terminal"), "terminal", "Util.spawnCommandLine('gnome-terminal')", this.menu, this.hoverIcon, false);
        this.help = new TextBoxItem(_("Help"), "help", "Util.spawnCommandLine('yelp')", this.menu, this.hoverIcon, false);
        this.shutdown = new TextBoxItem("Shutdown", "system-shutdown", "Session.ShutdownRemote()", this.menu, this.hoverIcon, false); //关机按钮
        this.shutdownMenu = new ShutdownMenu(this.menu, this.hoverIcon);  //关机菜单

        this.shutdownBox.add_actor(this.shutdown.actor);
        this.shutdownBox.add_actor(this.shutdownMenu.actor);

        this.actor.add_actor(this.hoverIcon.actor);
        this.actor.add_actor(this.home.actor);
        this.actor.add_actor(this.documents.actor);
        this.actor.add_actor(this.pictures.actor);
        this.actor.add_actor(this.music.actor);
        this.actor.add_actor(this.videos.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.computer.actor);
        this.actor.add_actor(this.control.actor);
        //this.actor.add_actor(this.packageItem.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.run.actor);
        this.actor.add_actor(this.terminal.actor);
        this.actor.add_actor(this.help.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.shutdownBox);
        this.actor.add_actor(this.shutdownMenu.menu.actor);
    }
}

function FavoritesBox() {
    this._init();
}

FavoritesBox.prototype = {
    _init: function () {
        this.actor = new St.BoxLayout({
            vertical: true
        });
        this.actor._delegate = this;

        this._dragPlaceholder = null;
        this._dragPlaceholderPos = -1;
        this._animatingPlaceholdersCount = 0;
    },

    _clearDragPlaceholder: function () {
        if (this._dragPlaceholder) {
            this._dragPlaceholder.animateOutAndDestroy();
            this._dragPlaceholder = null;
            this._dragPlaceholderPos = -1;
        }
    },

    handleDragOver: function (source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed() || (!(source instanceof FavoritesButton) && app.get_id() in AppFavorites.getAppFavorites().getFavoriteMap())) return DND.DragMotionResult.NO_DROP;

        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let numFavorites = favorites.length;

        let favPos = favorites.indexOf(app);

        let children = this.actor.get_children();
        let numChildren = children.length;
        let boxHeight = this.actor.height;

        // Keep the placeholder out of the index calculation; assuming that
        // the remove target has the same size as "normal" items, we don't
        // need to do the same adjustment there.
        if (this._dragPlaceholder) {
            boxHeight -= this._dragPlaceholder.actor.height;
            numChildren--;
        }

        let pos = Math.round(y * numFavorites / boxHeight);

        if (pos != this._dragPlaceholderPos && pos <= numFavorites) {
            if (this._animatingPlaceholdersCount > 0) {
                let appChildren = children.filter(function (actor) {
                    return (actor._delegate instanceof FavoritesButton);
                });
                this._dragPlaceholderPos = children.indexOf(appChildren[pos]);
            } else {
                this._dragPlaceholderPos = pos;
            }

            // Don't allow positioning before or after self
            if (favPos != -1 && (pos == favPos || pos == favPos + 1)) {
                if (this._dragPlaceholder) {
                    this._dragPlaceholder.animateOutAndDestroy();
                    this._animatingPlaceholdersCount++;
                    this._dragPlaceholder.actor.connect('destroy', Lang.bind(this, function () {
                        this._animatingPlaceholdersCount--;
                    }));
                }
                this._dragPlaceholder = null;

                return DND.DragMotionResult.CONTINUE;
            }

            // If the placeholder already exists, we just move
            // it, but if we are adding it, expand its size in
            // an animation
            let fadeIn;
            if (this._dragPlaceholder) {
                this._dragPlaceholder.actor.destroy();
                fadeIn = false;
            } else {
                fadeIn = true;
            }

            this._dragPlaceholder = new DND.GenericDragPlaceholderItem();
            this._dragPlaceholder.child.set_width(source.actor.height);
            this._dragPlaceholder.child.set_height(source.actor.height);
            this.actor.insert_actor(this._dragPlaceholder.actor, this._dragPlaceholderPos);
            if (fadeIn) this._dragPlaceholder.animateIn();
        }

        let srcIsFavorite = (favPos != -1);

        if (srcIsFavorite) return DND.DragMotionResult.MOVE_DROP;

        return DND.DragMotionResult.COPY_DROP;
    },

    // Draggable target interface
    acceptDrop: function (source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed()) {
            return false;
        }

        let id = app.get_id();

        let favorites = AppFavorites.getAppFavorites().getFavoriteMap();

        let srcIsFavorite = (id in favorites);

        let favPos = 0;
        let children = this.actor.get_children();
        for (let i = 0; i < this._dragPlaceholderPos; i++) {
            if (this._dragPlaceholder && children[i] == this._dragPlaceholder.actor) continue;

            if (!(children[i]._delegate instanceof FavoritesButton)) continue;

            let childId = children[i]._delegate.app.get_id();
            if (childId == id) continue;
            if (childId in favorites) favPos++;
        }

        Meta.later_add(Meta.LaterType.BEFORE_REDRAW, Lang.bind(this, function () {
            let appFavorites = AppFavorites.getAppFavorites();
            if (srcIsFavorite) appFavorites.moveFavoriteToPos(id, favPos);
            else appFavorites.addFavoriteAtPos(id, favPos);
            return false;
        }));

        return true;
    }
}

/*****************************  MyApplet  **************************/
function MyApplet(orientation, panel_height, instance_id) {
    this._init(orientation, panel_height, instance_id);
}

MyApplet.prototype = {
    __proto__: Applet.TextIconApplet.prototype,

    _init: function (orientation, panel_height, instance_id) {
        Applet.TextIconApplet.prototype._init.call(this, orientation, panel_height, instance_id);

        try {
            this.settings = new Settings.AppletSettings(this, "menu@cdos.org", instance_id);           
            this.set_applet_tooltip(_("Menu"));

            this.menuManager = new PopupMenu.PopupMenuManager(this);
            this.menu = new Applet.AppletPopupMenu(this, orientation);
            this.menuManager.addMenu(this.menu);

            this.actor.connect('key-press-event', Lang.bind(this, this._onSourceKeyPress));

            this.menu.actor.add_style_class_name('menu-background');
            this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));

            this._updateIcon();

            global.settings.connect("changed::menu-icon", Lang.bind(this, function () {
                this._updateIcon();
            }));

            this.set_applet_label(_("Menu"));
            /*
            let menuLabel = global.settings.get_string("menu-text");
            if (menuLabel !="" && menuLabel != "Menu") {
                this.set_applet_label(_(menuLabel));
            }
            global.settings.connect("changed::menu-text", Lang.bind(this, function () {
                this.set_applet_label(global.settings.get_string("menu-text"));
            }));
            */
            this._searchInactiveIcon = new St.Icon({
                style_class: 'menu-search-entry-icon',
                icon_name: 'edit-find',
                icon_type: St.IconType.SYMBOLIC
            });
            this._searchActiveIcon = new St.Icon({
                style_class: 'menu-search-entry-icon',
                icon_name: 'edit-clear',
                icon_type: St.IconType.SYMBOLIC
            });
            this._searchTimeoutId = 0;
            this._searchIconClickedId = 0;
            this._applicationsButtons = new Array();
            this._favoritesButtons = new Array();
            this._selectedItemIndex = null;
            this._previousSelectedItemIndex = null;
            this._activeContainer = null;

            this._display();
            appsys.connect('installed-changed', Lang.bind(this, this._refreshApps));
            AppFavorites.getAppFavorites().connect('changed', Lang.bind(this, this._refreshFavs));

            this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateToggled));


            this.hover_delay = global.settings.get_int("menu-hover-delay") / 1000;
            global.settings.connect("changed::menu-hover-delay", Lang.bind(this, function () {
                this.hover_delay = global.settings.get_int("menu-hover-delay") / 1000;
            }));
            /*
            global.display.connect('overlay-key', Lang.bind(this, function () {
                try {
                    this.menu.toggle();
                } catch (e) {
                    global.logError(e);
                }
            }));
            */
            this.edit_menu_item = new Applet.MenuItem(_("Edit menu"), Gtk.STOCK_EDIT, Lang.bind(this, this._launch_editor));
            this._applet_context_menu.addMenuItem(this.edit_menu_item);

        } catch (e) {
            global.logError(e);
        }
    },

    on_orientation_changed: function (orientation) {     //改变方向
        this.menu.destroy();
        this.menu = new Applet.AppletPopupMenu(this, orientation);
        this.menuManager.addMenu(this.menu);

        this.menu.actor.add_style_class_name('menu-background');
        this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateChanged));
        this.menu.connect('open-state-changed', Lang.bind(this, this._onOpenStateToggled));

        this._display();
    },

    _launch_editor: function () {
        Util.spawnCommandLine("cdos-menu-editor");
    },

    on_applet_clicked: function (event) {
        this.menu.toggle();
    },

    _onSourceKeyPress: function (actor, event) {
        let symbol = event.get_key_symbol();

        if (symbol == Clutter.KEY_space || symbol == Clutter.KEY_Return) {  //按下空格或回车
            this.menu.toggle();  //切换菜单
            return true;
        } else if (symbol == Clutter.KEY_Escape && this.menu.isOpen) {   //按下Esc
            this.menu.close();
            return true;
        } else if (symbol == Clutter.KEY_Down) {   //按下下键
            if (!this.menu.isOpen) this.menu.toggle();
            this.menu.actor.navigate_focus(this.actor, Gtk.DirectionType.DOWN, false);
            return true;
        } else return false;
    },

    _onOpenStateChanged: function (menu, open) {
        if (open) {
            this.actor.add_style_pseudo_class('active');
        } else {
            this.actor.remove_style_pseudo_class('active');
            if (this.searchActive) {
                this.resetSearch();
                this._select_category(null, this._allAppsCategoryButton);
            }
        }
    },

    destroy: function () {
        this.actor._delegate = null;
        this.menu.destroy();
        this.actor.destroy();
        this.emit('destroy');
    },

    _updateIcon: function () {
        let icon_file = global.settings.get_string("menu-icon");
        if(!icon_file) {
            icon_file = global.datadir + '/theme/menu.svg';
        }
        if(!GLib.file_test(icon_file, GLib.FileTest.EXISTS))
            icon_file = global.datadir + '/theme/menu.png';
        try {
            this.set_applet_icon_path(icon_file);
        } catch (e) {
            global.log("WARNING : Could not load icon file \"" + icon_file + "\" for menu button");
        }
    },

    _onMenuKeyPress: function (actor, event) {

        let symbol = event.get_key_symbol();

        if (symbol == Clutter.KEY_Super_L && this.menu.isOpen) {
            this.menu.close();
            return true;
        }

        if (this._activeContainer === null && symbol == Clutter.KEY_Up) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            this._selectedItemIndex = children.length;
        } else if (this._activeContainer === null && symbol == Clutter.KEY_Down) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            this._selectedItemIndex = -1;
        } else if (this._activeContainer === null) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = -1;
            this._previousSelectedItemIndex = -1;
        }


        let children = this._activeContainer.get_children();

        if (children.length == 0) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = -1;
            this._previousSelectedItemIndex = -1;
            children = this._activeContainer.get_children();
        }

        let index = this._selectedItemIndex;

        if (symbol == Clutter.KEY_Up) {
            if (this._activeContainer == this.categoriesBox) index = this._selectedItemIndex - 1 < 0 ? 0 : this._selectedItemIndex - 2;
            else index = this._selectedItemIndex - 1 < 0 ? 0 : this._selectedItemIndex - 1;
        } else if (symbol == Clutter.KEY_Down) {
            if (this._activeContainer == this.categoriesBox && this._selectedItemIndex != -1) index = this._selectedItemIndex + 2 >= children.length ? children.length - 2 : this._selectedItemIndex + 2;
            else index = this._selectedItemIndex + 1 == children.length ? children.length - 1 : this._selectedItemIndex + 1;
        } else if (symbol == Clutter.KEY_Right && this._activeContainer === this.categoriesBox) {
            this._activeContainer = this.categoriesBox;
            children = this._activeContainer.get_children();
            index = 0;
            this._previousSelectedItemIndex = this._selectedItemIndex;
            this._selectedItemIndex = -1;
        } else {
            return false;
        }

        if (index == this._selectedItemIndex) {
            return true;
        }

        if (this._activeContainer == this.categoriesBox) {
            if (index >= children.length - 1) index = children.length - 2;
        } else {
            if (index >= children.length) index = children.length - 1;
        }

        this._selectedItemIndex = index;
        let item_actor = children[this._selectedItemIndex];

        if (!item_actor || item_actor === this.searchEntry) {
            return false;
        }

        item_actor._delegate.emit('enter-event');
        return true;
    },

    _addEnterEvent: function (button, callback) {
        let _callback = Lang.bind(this, function () {
            let parent = button.actor.get_parent();
            if (this._activeContainer === this.categoriesBox && parent !== this._activeContainer) {
                this._previousSelectedItemIndex = this._selectedItemIndex;
            }
            this._activeContainer = parent;
            let children = this._activeContainer.get_children();
            for (let i = 0, l = children.length; i < l; i++) {
                if (button.actor === children[i]) {
                    this._selectedItemIndex = i;
                }
            };
            callback();
        });
        button.connect('enter-event', _callback);
        button.actor.connect('enter-event', _callback);
    },

    _clearSelections: function (container) {
        container.get_children().forEach(function (actor) {
            actor.style_class = "menu-category-button";
            if (actor._delegate && actor._delegate.menu)
                actor._delegate.menu.close(true);
        });
    },

    _onOpenStateToggled: function (menu, open) {
        if (open) {
            global.stage.set_key_focus(this.searchEntry);
            this._selectedItemIndex = null;
            this._activeContainer = null;
            let monitorHeight = Main.layoutManager.primaryMonitor.height;
            this._appletStyles();
        } else {
            this.closeApplicationsContextMenus(null, false);
            //this.resetSearch();
            //this._clearSelections(this.categoriesBox);
        }
    },

    _appletStyles: function () {   //设置小程序的尺寸
            this.favsBox.style = "min-height: 455px;min-width: 235px;";
            this.favBoxWith = (this.favsBox.get_allocation_box().x2 - this.favsBox.get_allocation_box().x1);
            let searchEntryWidth = this.favBoxWith - this._searchInactiveIcon.get_theme_node().get_length('icon-size')
            let scrollBoxHeight = (this.favsBox.get_allocation_box().y2 - this.favsBox.get_allocation_box().y1) - (this.selectedAppBox.get_allocation_box().y2 - this.selectedAppBox.get_allocation_box().y1);  

            this.searchEntry.style = "width:" + searchEntryWidth + "px";
            this.categoriesScrollBox.set_style("width: " + this.favBoxWith + "px; height: " + scrollBoxHeight + "px;");
    },

    _refreshApps: function () {
        this._applicationsButtons = new Array();

        //Remove all categories
        this.categoriesBox.get_children().forEach(Lang.bind(this, function (child) {
            child.destroy();
        }));

        this._allAppsCategoryButton = new CategoryButton(null);
        this._allAppsCategoryButton.actor.connect('button-press-event', Lang.bind(this, function () {
            this._categoryButtonsHidden(false);
            if (!this._allAppsCategoryButton.menu.isOpen)
                this._select_category(null, this._allAppsCategoryButton);
        }));
        this._addEnterEvent(this._allAppsCategoryButton, Lang.bind(this, function () {
            if (!this.searchActive) {
                this._allAppsCategoryButton.isHovered = true;
                Tweener.addTween(this, {
                    time: this.hover_delay,
                    onComplete: function () {
                        if (this._allAppsCategoryButton.isHovered) {
                            //this._select_category(null, this._allAppsCategoryButton);
                            this._allAppsCategoryButton.actor.add_style_class_name("menu-category-button-selected");
                        }
                    }
                });
            }
        }));
        this._allAppsCategoryButton.actor.connect('leave-event', Lang.bind(this, function () {
            this._allAppsCategoryButton.isHovered = false;
            if (!this._allAppsCategoryButton.menu.isOpen)
                this._allAppsCategoryButton.actor.remove_style_class_name("menu-category-button-selected");
        }));

        this.categoriesBox.add_actor(this._allAppsCategoryButton.actor);
        this.categoriesBox.add_actor(this._allAppsCategoryButton.menu.actor);

        //let trees = [appsys.get_tree(), appsys.get_settings_tree()];
        let trees = [appsys.get_tree()];
        for (let i in trees) {
            let tree = trees[i];
            let root = tree.get_root_directory();

            let iter = root.iter();
            let nextType;
            while ((nextType = iter.next()) != CMenu.TreeItemType.INVALID) {
                if (nextType == CMenu.TreeItemType.DIRECTORY) {
                    let dir = iter.get_directory();
                    if (dir.get_is_nodisplay()) continue;
                    this.applicationsByCategory[dir.get_menu_id()] = new Array();
                    this._loadCategory(dir);
                    if (this.applicationsByCategory[dir.get_menu_id()].length > 0) {
                        let categoryButton = new CategoryButton(dir);
                        categoryButton.actor.connect('button-press-event', Lang.bind(this, function () {
                            if (!categoryButton.menu.isOpen) {
                                this._select_category(dir, categoryButton);
                            }
                        }));
                        this._addEnterEvent(categoryButton, Lang.bind(this, function () {
                            if (!this.searchActive) {
                                categoryButton.isHovered = true;
                                Tweener.addTween(this, {
                                    time: this.hover_delay,
                                    onComplete: function () {
                                        if (categoryButton.isHovered) {
                                            //this._select_category(dir, categoryButton);
                                            categoryButton.actor.add_style_class_name("menu-category-button-selected");
                                        }
                                    }
                                });
                            }
                        }));
                        categoryButton.actor.connect('leave-event', function () {
                            categoryButton.isHovered = false;
                            if (!categoryButton.menu.isOpen)
                                categoryButton.actor.remove_style_class_name("menu-category-button-selected");
                        });
                        this.categoriesBox.add_actor(categoryButton.actor);
                        this.categoriesBox.add_actor(categoryButton.menu.actor);
                    }
                }
            }
        }

        this._select_category(null, this._allAppsCategoryButton);
    },

    _refreshFavs: function () {
        //Remove all favorites
    	this.favsBox.get_children().forEach(Lang.bind(this, function (child) {
            child.destroy();
        }));

        let favoritesBox = new FavoritesBox();
        this.favsBox.add_actor(favoritesBox.actor, {
            y_align: St.Align.END,
            y_fill: false
        });

        //Load favorites again
        this._favoritesButtons = new Array();
        let launchers = global.settings.get_strv('favorite-apps');
        let appSys = Cinnamon.AppSystem.get_default();
        let j = 0;
        for (let i = 0; i < launchers.length; ++i) {
            let app = appSys.lookup_app(launchers[i]);
            if (!app) app = appSys.lookup_settings_app(launchers[i]);
            if (app) {
                let button = new FavoritesButton(this, app, launchers.length); // + 3 because we're adding 3 system buttons at the bottom
                this._favoritesButtons[app] = button;
                favoritesBox.actor.add_actor(button.actor, {
                    y_align: St.Align.END,
                    y_fill: false
                });
                favoritesBox.actor.add_actor(button.menu.actor, {
                    y_align: St.Align.END,
                    y_fill: false
                });
                button.actor.connect('enter-event', Lang.bind(this, function () {
                    this.selectedAppTitle.set_text(button.app.get_name());
                    if (button.app.get_description()) this.selectedAppDescription.set_text(button.app.get_description());
                    else this.selectedAppDescription.set_text("");
                }));
                button.actor.connect('leave-event', Lang.bind(this, function () {
                    this.selectedAppTitle.set_text("");
                    this.selectedAppDescription.set_text("");
                }));
                ++j;
            }
        }
    },

    _loadCategory: function (dir, top_dir) {
        var iter = dir.iter();
        var nextType;
        if (!top_dir) top_dir = dir;
        while ((nextType = iter.next()) != CMenu.TreeItemType.INVALID) {
            if (nextType == CMenu.TreeItemType.ENTRY) {
                var entry = iter.get_entry();
                if (!entry.get_app_info().get_nodisplay()) {
                    var app = appsys.lookup_app_by_tree_entry(entry);
                    if (!app) app = appsys.lookup_settings_app_by_tree_entry(entry);
                    if (!this.applicationsByCategory[top_dir.get_menu_id()]) this.applicationsByCategory[top_dir.get_menu_id()] = new Array();
                    this.applicationsByCategory[top_dir.get_menu_id()].push(app);
                }
            } else if (nextType == CMenu.TreeItemType.DIRECTORY) {
                this._loadCategory(iter.get_directory(), top_dir);
            }
        }
    },

    _display: function () {
        this._activeContainer = null;
        let section = new PopupMenu.PopupMenuSection();
        this.menu.addMenuItem(section);

        this.leftPane = new St.Bin();

        this.favsBox = new St.BoxLayout({
            vertical: true
        });

        this.selectedAppBox = new St.BoxLayout({
            style_class: 'menu-selected-app-box',
            vertical: true
        });
        this.selectedAppBox.hide();
        this.selectedAppTitle = new St.Label({
            style_class: 'menu-selected-app-title',
            text: ""
        });
        this.selectedAppDescription = new St.Label({
            style_class: 'menu-selected-app-description',
            text: ""
        });
        this.categoriesBox = new St.BoxLayout({
            vertical: true
        });
        this.categoriesScrollBox = new St.ScrollView({
            x_fill: true, 
            y_fill: false, 
            y_align: St.Align.START, 
            style_class: 'vfade menu-applications-scrollbox'
        });

        let vscroll = this.categoriesScrollBox.get_vscroll_bar();
        vscroll.connect('scroll-start', Lang.bind(this, function () {
            this.menu.passEvents = true;
        }));
        vscroll.connect('scroll-stop', Lang.bind(this, function () {
            this.menu.passEvents = false;
        }));
        this.categoriesScrollBox.set_policy(Gtk.PolicyType.NEVER, Gtk.PolicyType.AUTOMATIC);

        this.mainBox = new St.BoxLayout({
            style_class: 'menu-applications-box',
            vertical: false
        });

        this.leftPaneBox = new St.BoxLayout({
            style_class: 'menu-favorites-box',
            vertical: true
        });
        this.leftPaneBox.style = "padding: 0px";

        this.appsButtonSearchBox = new St.BoxLayout({
            style_class: 'menu-favorites-box',
            vertical: true
        });
        this.appsButtonSearchBox.style = "padding: 0px";
        
        this.rightButtonsBox = new RightButtonsBox(this, this.menu);   //右边栏

        this.appsButton = new AllProgramsItem(_("All Programs"), "forward", this, false);  //所有程序按钮
        /*
        this.appsButton = new St.BoxLayout({
            style_class: 'button-changer'
        });
        this.appsButton.add_actor(this._appsButton.actor);        
        
        this.appsButton.set_style("padding-right: 0px;padding-left: 0px");        
        */
        this.searchBox = new St.BoxLayout({      //searchBox
            style_class: 'menu-search-box'
        });
        this.searchBox.set_style("padding-right: 0px;padding-left: 0px");
        this.searchEntry = new St.Entry({    //searchEntry
            name: 'menu-search-entry',
            style_class: 'menu-search-entry',
            hint_text: _("Type to search..."),
            track_hover: true,
            can_focus: true
        });
        this.searchEntry.set_secondary_icon(this._searchInactiveIcon);
        this.searchActive = false;
        this.searchEntryText = this.searchEntry.clutter_text;
        this.searchEntryText.connect('text-changed', Lang.bind(this, this._onSearchTextChanged));
        this.searchEntryText.connect('key-press-event', Lang.bind(this, this._onMenuKeyPress));
        this._previousSearchPattern = "";

        this.applicationsByCategory = {};
        this._refreshApps();
        this._refreshFavs();

        this.placesButton = new PlaceCategoryButton();
        this.placesButton.actor.connect('button-press-event', Lang.bind(this, function () {
            if (!this.placesButton.menu.isOpen) {
                this._select_places(this.placesButton);
            }
        }));
        this._addEnterEvent(this.placesButton, Lang.bind(this, function () {
            if (!this.searchActive) {
                this.placesButton.isHovered = true;
                Tweener.addTween(this, {
                    time: this.hover_delay,
                    onComplete: function () {
                        if (this.placesButton.isHovered) {
                            //this._select_places(this.placesButton);
                            this.placesButton.actor.add_style_class_name("menu-category-button-selected");                                  
                        }
                    }
                });
            }
        }));
        this.placesButton.actor.connect('leave-event', Lang.bind(this, function () {
            this.placesButton.isHovered = false;
            this.placesButton.actor.remove_style_class_name("menu-category-button-selected");
        }));
       
        this.leftPane.set_child(this.favsBox, {
            y_align: St.Align.END,
            y_fill: false
        });
       
        this.selectedAppBox.add_actor(this.selectedAppTitle);
        this.selectedAppBox.add_actor(this.selectedAppDescription);
        this.categoriesBox.add_actor(this.placesButton.actor);
        this.categoriesBox.add_actor(this.placesButton.menu.actor);
        this.categoriesScrollBox.add_actor(this.categoriesBox);
        this.searchBox.add_actor(this.searchEntry);
        this.appsButtonSearchBox.add_actor(this.appsButton.actor);
        //this.appsButtonSearchBox.add_actor(this.appsButton);        
        this.appsButtonSearchBox.add_actor(this.searchBox);
        this.leftPaneBox.add_actor(this.selectedAppBox);
        this.leftPaneBox.add_actor(this.leftPane);
        this.leftPaneBox.add_actor(this.appsButtonSearchBox);
        this.mainBox.add_actor(this.leftPaneBox);
        this.mainBox.add_actor(this.rightButtonsBox.actor);
        section.actor.add_actor(this.mainBox);
    },

    _clearApplicationsBox: function (selectedActor) {
        let actors = this.categoriesBox.get_children();

        for (let i = 0; i < actors.length; i++) {
            let actor = actors[i];
            if (actor._delegate && actor._delegate.menu)
                actor._delegate.removeItems();
        }
    },

    _select_category: function (dir, categoryButton) {
        this.resetSearch();
        this._clearApplicationsBox(categoryButton.actor);
        if (dir) this._displayButtons(this._listApplications(dir.get_menu_id()), null, categoryButton);
        else this._displayButtons(this._listApplications(null), null, categoryButton);
        this.closeApplicationsContextMenus(null, false);
    },

    closeApplicationsContextMenus: function (excludeApp, animate) {
        for (var app in this._applicationsButtons) {
            if (app != excludeApp && this._applicationsButtons[app].menu.isOpen) {
                if (animate) this._applicationsButtons[app].toggleMenu();
                else this._applicationsButtons[app].closeMenu();
            }
        }

        for (var fav in this._favoritesButtons) {
            if (fav != excludeApp && this._favoritesButtons[fav].menu.isOpen) {
                if (animate) this._favoritesButtons[fav].toggleMenu();
                else this._favoritesButtons[fav].closeMenu();
            }
        }
    },

    _displayButtons: function (apps, places, categoryButton) {
        if (apps) {
            for (var i = 0; i < apps.length; i++) {
                let app = apps[i];
                if (!this._applicationsButtons[app]) {
                    let applicationButton = new ApplicationButton(this, app);
                    applicationButton.actor.connect('leave-event', Lang.bind(this, function () {
                        this.selectedAppTitle.set_text("");
                        this.selectedAppDescription.set_text("");
                        applicationButton.actor.remove_style_class_name("menu-category-button-selected");
                    }));
                    this._addEnterEvent(applicationButton, Lang.bind(this, function () {
                        this.selectedAppTitle.set_text(applicationButton.app.get_name());
                        if (applicationButton.app.get_description()) this.selectedAppDescription.set_text(applicationButton.app.get_description());
                        else this.selectedAppDescription.set_text("");
                        //this._clearSelections(this.categoriesBox);
                        applicationButton.actor.add_style_class_name("menu-category-button-selected");
                    }));
                    this._applicationsButtons[app] = applicationButton;
                }
                categoryButton.menu.addMenuItem(this._applicationsButtons[app]);
            }
        }

        if (places) {
            for (var i = 0; i < places.length; i++) {
                let place = places[i];
                let button = new PlaceButton(this, place, place.name);
                this._addEnterEvent(button, Lang.bind(this, function () {
                    //this._clearSelections(this.categoriesBox);
                    button.actor.add_style_class_name("menu-category-button-selected");
                }));
                button.actor.connect('leave-event', Lang.bind(this, function () {
                    button.actor.remove_style_class_name("menu-category-button-selected");
                }));
                categoryButton.menu.box.add_actor(button.actor);
            }
        }
    },

    _select_places: function (button) {
        this.resetSearch();
        this._clearApplicationsBox(button.actor);

        let bookmarks = this._listBookmarks();
        let devices = this._listDevices();
        this._displayButtons(null, bookmarks.concat(devices), button);
    },

    _categoryButtonsHidden: function (hide) {
        let categoriesBoxChilren = this.categoriesBox.get_children();
        if (this.categorysHidden && !hide) {
            for (var i = 0; i < categoriesBoxChilren.length; i++) {
                let item_actor = categoriesBoxChilren[i];
                item_actor.show();
            }
            this.categorysHidden = false;
            this._allAppsCategoryButton.label.set_text(_("All Applications"));
        }else if (hide) {
            for (var i = 0; i < categoriesBoxChilren.length; i++) {
                let item_actor = categoriesBoxChilren[i];
                if (item_actor != this._allAppsCategoryButton.actor && item_actor != this._allAppsCategoryButton.menu.actor)
                    item_actor.hide();
            }
            this.categorysHidden = true;
            this._allAppsCategoryButton.label.set_text(_("Back to Categorys"));
        }

    },

    resetSearch: function () {
        this.searchEntry.set_text("");
        this.searchActive = false;
        global.stage.set_key_focus(this.searchEntry);
    },

    _onSearchTextChanged: function (se, prop) {
        this._clearSelections(this.categoriesBox);
        this.searchActive = this.searchEntry.get_text() != '';
        if (this.searchActive) {
            this.searchEntry.set_secondary_icon(this._searchActiveIcon);

            if (this._searchIconClickedId == 0) {
                this._searchIconClickedId = this.searchEntry.connect('secondary-icon-clicked', Lang.bind(this, function () {
                    this.resetSearch();
                    this._select_category(null, this._allAppsCategoryButton);
                }));
            }
        } else {
            if (this._searchIconClickedId > 0) this.searchEntry.disconnect(this._searchIconClickedId);
            this._searchIconClickedId = 0;

            this.searchEntry.set_secondary_icon(this._searchInactiveIcon);
        }
        if (!this.searchActive) {
            if (this._searchTimeoutId > 0) {
                Mainloop.source_remove(this._searchTimeoutId);
                this._searchTimeoutId = 0;
            }
            this._categoryButtonsHidden(false);
            return;
        }
        if (this._searchTimeoutId > 0) return;
        this._searchTimeoutId = Mainloop.timeout_add(150, Lang.bind(this, this._doSearch));
    },

    _listBookmarks: function (pattern) {
        let bookmarks = Main.placesManager.getBookmarks();
        var res = new Array();
        for (let id = 0; id < bookmarks.length; id++) {
            if (!pattern || bookmarks[id].name.toLowerCase().indexOf(pattern) != -1) res.push(bookmarks[id]);
        }
        return res;
    },

    _listDevices: function (pattern) {
        let devices = Main.placesManager.getMounts();
        var res = new Array();
        for (let id = 0; id < devices.length; id++) {
            if (!pattern || devices[id].name.toLowerCase().indexOf(pattern) != -1) res.push(devices[id]);
        }
        return res;
    },

    _listApplications: function (category_menu_id, pattern) {
        var applist;
        if (category_menu_id) applist = this.applicationsByCategory[category_menu_id];
        else {
            applist = new Array();
            for (directory in this.applicationsByCategory) applist = applist.concat(this.applicationsByCategory[directory]);
        }

        var res;
        if (pattern) {
            res = new Array();
            for (var i in applist) {
                let app = applist[i];
                if (app.get_name().toLowerCase().indexOf(pattern) != -1 || (app.get_description() && app.get_description().toLowerCase().indexOf(pattern) != -1) || (app.get_id() && app.get_id().slice(0, -8).toLowerCase().indexOf(pattern) != -1)) res.push(app);
            }
        } else res = applist;

        res.sort(function (a, b) {
            return a.get_name().toLowerCase() > b.get_name().toLowerCase();
        });

        return res;
    },

    _doSearch: function () {
        if (this.leftPane.get_child() === this.favsBox && this.menu.isOpen) {
            this.leftPane.set_child(this.categoriesScrollBox, {
                y_align: St.Align.END,
                y_fill: false
            });
            this.appsButton.label.set_text(' Favorites');
            this.selectedAppBox.show();
            this._appletStyles();
        }
        this._searchTimeoutId = 0;
        let pattern = this.searchEntryText.get_text().replace(/^\s+/g, '').replace(/\s+$/g, '').toLowerCase();
        if (pattern == this._previousSearchPattern) return false;
        this._previousSearchPattern = pattern;

        this._activeContainer = null;
        this._selectedItemIndex = null;
        this._previousSelectedItemIndex = null;

        // _listApplications returns all the applications when the search
        // string is zero length. This will happend if you type a space
        // in the search entry.
        if (pattern.length == 0) {
            return false;
        }

        var appResults = this._listApplications(null, pattern);

        var placesResults = new Array();

        var bookmarks = this._listBookmarks(pattern);
        for (var i in bookmarks) placesResults.push(bookmarks[i]);

        var devices = this._listDevices(pattern);
        for (var i in devices) placesResults.push(devices[i]);

        this._clearApplicationsBox();
        this._displayButtons(appResults, placesResults, this._allAppsCategoryButton);
        this._allAppsCategoryButton.menu.toggle();
        

        let categoriesBoxChilren = this.categoriesBox.get_children();
        if (categoriesBoxChilren.length > 0) {
            this._activeContainer = this.categoriesBox;
            this._selectedItemIndex = 0;
            let item_actor = categoriesBoxChilren[this._selectedItemIndex];
            if (item_actor && item_actor !== this.searchEntry && item_actor._delegate) {
                item_actor._delegate.emit('enter-event');
            }
        }
        this._categoryButtonsHidden(true);
            

        return false;
    }


};

function main(metadata, orientation, panel_height, instance_id) {  
    let myApplet = new MyApplet(orientation, panel_height, instance_id);
    return myApplet;      
}
