const St = imports.gi.St;
const Lang = imports.lang;
const PopupMenu = imports.ui.popupMenu;
const AppFavorites = imports.ui.appFavorites;
const GnomeSession = imports.misc.gnomeSession;
const Cinnamon = imports.gi.Cinnamon;

const ScreenSaver = imports.misc.screenSaver;
const DND = imports.ui.dnd;
const Meta = imports.gi.Meta;
const Clutter = imports.gi.Clutter;
const Main = imports.ui.main;
const PopupMenu = imports.ui.popupMenu;
const AppFavorites = imports.ui.appFavorites;
const ScreenSaver = imports.misc.screenSaver;
const FileUtils = imports.misc.fileUtils;
const Tweener = imports.ui.tweener;
const Signals = imports.signals;
const Util = imports.misc.util;
const DND = imports.ui.dnd;
const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const AccountsService = imports.gi.AccountsService;

const ICON_SIZE = 16;
const MAX_FAV_ICON_SIZE = 32;
const CATEGORY_ICON_SIZE = 32;
const APPLICATION_ICON_SIZE = 16;
const HOVER_ICON_SIZE =48;

const USER_DESKTOP_PATH = FileUtils.getUserDesktopDir();

var Session = new GnomeSession.SessionManager();
const AppletDir = imports.ui.appletManager.applets['menu@cdos.org'];
const MainApplet = AppletDir.applet;
//const Items = AppletDir.items;


function ApplicationContextMenuItem(appButton, label, action) {
    this._init(appButton, label, action);
}

ApplicationContextMenuItem.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function (appButton, label, action) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            focusOnHover: false
        });

        this._appButton = appButton;
        this._action = action;
        this.label = new St.Label({
            text: label
        });
        this.addActor(this.label);
    },

    activate: function (event) {
        this._appButton.toggleMenu();
        switch (this._action) {
        case "add_to_panel":
            let winListApplet = false;
            try{
                winListApplet = imports.ui.appletManager.applets['window-list@cdos.org'];
            }catch (e) {
            }
            if (winListApplet)
                winListApplet.applet.GetAppFavorites().addFavorite(this._appButton.app.get_id());
            else{
                let settings = new Gio.Settings({
                    schema: 'org.cdos'
                });
                let desktopFiles = settings.get_strv('panel-launchers');
                desktopFiles.push(this._appButton.app.get_id());
                settings.set_strv('panel-launchers', desktopFiles);
            }
            break;
        case "add_to_desktop":
            let file = Gio.file_new_for_path(this._appButton.app.get_app_info().get_filename());
            let destFile = Gio.file_new_for_path(USER_DESKTOP_PATH + "/" + this._appButton.app.get_id());
            try {
                file.copy(destFile, 0, null, function () {});
                // Need to find a way to do that using the Gio library, but modifying the access::can-execute attribute on the file object seems unsupported
                Util.spawnCommandLine("chmod +x \"" + USER_DESKTOP_PATH + "/" + this._appButton.app.get_id() + "\"");
            } catch (e) {
                global.log(e);
            }
            break;
        case "add_to_favorites":
            AppFavorites.getAppFavorites().addFavorite(this._appButton.app.get_id());
            break;
        case "remove_from_favorites":
            AppFavorites.getAppFavorites().removeFavorite(this._appButton.app.get_id());
            break;
        }
        return false;
    }
};

function GenericApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

GenericApplicationButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (appsMenuButton, app, withMenu) {
        this.app = app;
        this.appsMenuButton = appsMenuButton;
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });

        this.withMenu = withMenu;

        if (this.withMenu) {
            this.menu = new ApplicationMenu(this.actor);
            this.menu.actor.set_style_class_name('menu-context-menu');
        }
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
        if (event.get_button() == 2) {
            this.appsMenuButton._clearSelections(this.appsMenuButton.categoriesBox);
        }
        if (event.get_button() == 3) {
            this.appsMenuButton.closeApplicationsContextMenus(this.app, true);
            this.toggleMenu();
        }
        return true;
    },

    activate: function (event) {
        this.app.open_new_window(-1);
        this.appsMenuButton.menu.close();
    },

    closeMenu: function () {
        if (this.withMenu) this.menu.close();
    },

    toggleMenu: function () {
        if (!this.withMenu) return;

        if (!this.menu.isOpen) {
            let children = this.menu.box.get_children();
            for (var i in children) {
                children[i].destroy();
            }
            let menuItem;
            menuItem = new ApplicationContextMenuItem(this, _("Add to panel"), "add_to_panel");
            this.menu.addMenuItem(menuItem);
            if (USER_DESKTOP_PATH) {
                menuItem = new ApplicationContextMenuItem(this, _("Add to desktop"), "add_to_desktop");
                this.menu.addMenuItem(menuItem);
            }
            if (AppFavorites.getAppFavorites().isFavorite(this.app.get_id())) {
                menuItem = new ApplicationContextMenuItem(this, _("Remove from favorites"), "remove_from_favorites");
                this.menu.addMenuItem(menuItem);
            } else {
                menuItem = new ApplicationContextMenuItem(this, _("Add to favorites"), "add_to_favorites");
                this.menu.addMenuItem(menuItem);
            }
        }
        this.menu.toggle();
    },
}

function ApplicationButton(appsMenuButton, app) {
    this._init(appsMenuButton, app);
}

ApplicationButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function (appsMenuButton, app) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);

        this.actor.add_style_class_name('menu-category-button');
        this.icon = this.app.create_icon_texture(APPLICATION_ICON_SIZE);
        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.icon);
        this.addActor(this.label);

        this._draggable = DND.makeDraggable(this.actor);
        this.isDraggableApp = true;
    },

    get_app_id: function () {
        return this.app.get_id();
    },

    getDragActor: function () {
        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let nbFavorites = favorites.length;
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let real_size = (0.7 * monitorHeight) / nbFavorites;
        let icon_size = 0.6 * real_size;
        if (icon_size > MAX_FAV_ICON_SIZE) icon_size = MAX_FAV_ICON_SIZE;
        return this.app.create_icon_texture(icon_size);
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function () {
        return this.actor;
    }
};
Signals.addSignalMethods(ApplicationButton.prototype);

function PlaceButton(appsMenuButton, place, button_name) {
    this._init(appsMenuButton, place, button_name);
}

PlaceButton.prototype = {
    __proto__: PopupMenu.PopupBaseMenuItem.prototype,

    _init: function (appsMenuButton, place, button_name) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.place = place;
        this.button_name = button_name;
        this.label = new St.Label({
            text: this.button_name,
            style_class: 'menu-application-button-label'
        });
        this.icon = place.iconFactory(APPLICATION_ICON_SIZE);
        this.addActor(this.icon);
        this.addActor(this.label);
    },

    activate: function () {
        this.place.launch();
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceButton.prototype);

function ApplicationMenu() {
    this._init.apply(this, arguments);
}

ApplicationMenu.prototype = {
    __proto__: PopupMenu.PopupMenuBase.prototype,

    _init: function (sourceActor) {
        PopupMenu.PopupMenuBase.prototype._init.call(this, sourceActor);

        this.actor = new St.BoxLayout({ style_class: 'popup-sub-menu'});

        this.actor.add_actor(this.box);
        this.actor._delegate = this;
        this.actor.clip_to_allocation = true;
        //this.actor.connect('key-press-event', Lang.bind(this, this._onKeyPressEvent));
        this.actor.hide();
    },

    open: function(animate) {
        if (this.isOpen)
            return;

        this.isOpen = true;

        this.actor.show();

        if (animate) {
            let [minHeight, naturalHeight] = this.actor.get_preferred_height(-1);
            this.actor.height = 0;
            Tweener.addTween(this.actor,
                             { height: naturalHeight,
                               time: 0.25,
                               onCompleteScope: this,
                               onComplete: function() {
                                   this.actor.set_height(-1);
                                   this.emit('open-state-changed', true);
                               }
                             });
        } else {
            this.emit('open-state-changed', true);
        }
    },

    close: function(animate) {
        if (!this.isOpen)
            return;

        this.isOpen = false;


        if (animate) {
            Tweener.addTween(this.actor,
                             { height: 0,
                               time: 0.25,
                               onCompleteScope: this,
                               onComplete: function() {
                                   this.actor.hide();
                                   this.actor.set_height(-1);
                                   this.emit('open-state-changed', false);
                               },
                             });
            } else {
                this.actor.hide();

                this.isOpen = false;
                this.emit('open-state-changed', false);
            }
    },

};

function CategoryButton(category) {
    this._init(category);
}

CategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (category) {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        var label;
        this.actor.add_style_class_name('menu-category-button');
        if (category) {
            let icon = category.get_icon();
            if (icon && icon.get_names) this.icon_name = icon.get_names().toString();
            else this.icon_name = "";
            label = category.get_name();
        } else label = _("All Applications");
        this.label = new St.Label({
            text: label,
            style_class: 'menu-category-button-label'
        });
        if (category && this.icon_name) {
            this.icon = new St.Icon({
                icon_name: this.icon_name,
                icon_size: CATEGORY_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            this.addActor(this.icon);
        }
        this.addActor(this.label);
        //this.actor.set_tooltip_text(category.get_name());
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.set_style_class_name('menu-context-menu');
    },

    activate: function () {
        this.menu.toggle();
    },

    removeItems: function () {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(CategoryButton.prototype);

function PlaceCategoryButton() {
    this._init();
}

PlaceCategoryButton.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function () {
        PopupMenu.PopupBaseMenuItem.prototype._init.call(this, {
            hover: false
        });
        this.actor.add_style_class_name('menu-category-button');
        this.label = new St.Label({
            text: _("Places"),
            style_class: 'menu-category-button-label'
        });
        this.icon = new St.Icon({
            icon_name: "folder",
            icon_size: CATEGORY_ICON_SIZE,
            icon_type: St.IconType.FULLCOLOR
        });
        this.addActor(this.icon);
        this.addActor(this.label);
        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.set_style_class_name('menu-context-menu');
    },

    activate: function () {
        this.menu.toggle();
    },

    removeItems: function () {
        let children = this.menu._getMenuItems();
        for (let i = 0; i < children.length; i++) {
            let item = children[i];
            this.menu.box.remove_actor(item.actor);
            if (item.menu)
                this.menu.box.remove_actor(item.menu.actor);
        }
    },

    destroy: function() {
        this.menu.destroy();
        PopupBaseMenuItem.prototype.destroy.call(this);
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1 | 2) {
            this.activate();
        }
        return true;
    }
};
Signals.addSignalMethods(PlaceCategoryButton.prototype);

function FavoritesButton(appsMenuButton, app, nbFavorites) {
    this._init(appsMenuButton, app, nbFavorites);
}

FavoritesButton.prototype = {
    __proto__: GenericApplicationButton.prototype,

    _init: function (appsMenuButton, app, nbFavorites) {
        GenericApplicationButton.prototype._init.call(this, appsMenuButton, app, true);
        let monitorHeight = Main.layoutManager.primaryMonitor.height;
        let real_size = (0.7 * monitorHeight) / nbFavorites;
        let icon_size = 0.6 * real_size;
        if (icon_size > MAX_FAV_ICON_SIZE) icon_size = MAX_FAV_ICON_SIZE;
        this.actor.style = "padding-top: " + (icon_size / 3) + "px;padding-bottom: " + (icon_size / 3) + "px; margin:auto;"

        this.actor.add_style_class_name('menu-favorites-button');
        this.addActor(app.create_icon_texture(icon_size));

        this.label = new St.Label({
            text: this.app.get_name(),
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.label);

        this._draggable = DND.makeDraggable(this.actor);
        this.isDraggableApp = true;
    },

    get_app_id: function () {
        return this.app.get_id();
    },

    getDragActor: function () {
        return new Clutter.Clone({
            source: this.actor
        });
    },

    // Returns the original actor that should align with the actor
    // we show as the item is being dragged.
    getDragActorSource: function () {
        return this.actor;
    }
};

function TextBoxItem(label, icon, func, parent, hoverIcon) {
    this._init(label, icon, func, parent, hoverIcon);
}

TextBoxItem.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (label, icon_name, func, parent, hoverIcon) {
        this.parent = parent;
        this.hoverIcon = hoverIcon;
        this.icon_name = icon_name;
	    this.icon = new St.Icon({style_class: 'popup-menu-icon', icon_type: St.IconType.FULLCOLOR, icon_name: icon_name, icon_size: ICON_SIZE });        
        this.func = func;
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, null);

        this.actor.add_style_class_name('menu-category-button');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);
        this.label = new St.Label({
            text: label,
            style_class: 'menu-application-button-label'
        });
        this.addActor(this.icon);
        this.addActor(this.label);
    },

    setActive: function (active) {
        if (active) { 
            this.actor.set_style_class_name('menu-category-button-selected');
            this.hoverIcon._refresh(this.icon_name);
        }
        else {
            this.actor.set_style_class_name('menu-category-button');
            this.hoverIcon._refresh();
        }
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            this.activate(event);
        }
    },

    activate: function (event) {
        eval(this.func);
        this.parent.close();
    }
};

function AllProgramsItem(label, icon, parent, rightP, leftP) {
    this._init(label, icon, parent, rightP, leftP);
}

AllProgramsItem.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (label, icon, parent) {
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, label);

        this.actor.set_style_class_name('menu-category-button');
        this.parent = parent;
        //this.actor.set_style_class_name('');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);
	    //this.label = new St.Label({text: label});
        this.label.realize();
	    this.icon = new St.Icon({style_class: 'popup-menu-icon', icon_type: St.IconType.FULLCOLOR, icon_name: icon, icon_size: ICON_SIZE });
	    this.addActor(this.icon);
	    this.addActor(this.label);
    },

    setActive: function(active){
        if (active)
            this.actor.add_style_class_name('menu-category-button-selected');
        else
            this.actor.remove_style_class_name('menu-category-button-selected');
    },

    _onButtonReleaseEvent: function(actor, event){
        if (event.get_button()==1){
            this.activate(event);
        }
    },

    activate: function(event){
        if (this.parent.leftPane.get_child() == this.parent.favsBox) {
            this.parent.leftPane.set_child(this.parent.categoriesScrollBox, {
                y_align: St.Align.END,
                y_fill: false
            });
            this.label.set_text(" Favorites");
            this.parent.selectedAppBox.show();
            this.parent._appletStyles();
        }else {
            this.parent.leftPane.set_child(this.parent.favsBox, {
                y_align: St.Align.END,
                y_fill: false
            });
            this.label.set_text(" All Programs");
            this.parent.selectedAppBox.hide();
        }
    }
};

function HoverIcon(menu) {
    this._init(menu);
}

HoverIcon.prototype = {
    _init: function (menu) {
        try {
            let userBox = new St.BoxLayout({ style_class: 'user-box', reactive: true, vertical: false });
            this._userIcon = new St.Bin({ style_class: 'user-icon'});
            userBox.connect('button-press-event', Lang.bind(this, function() {
                menu.toggle();
                Util.spawnCommandLine("cdos-settings user");
            }));

            this._userIcon.hide();
            userBox.add(this._userIcon,
                    { x_fill:  true,
                        y_fill:  false,
                        x_align: St.Align.END,
                        y_align: St.Align.START });
            this.userLabel = new St.Label(({ style_class: 'user-label'}));
            userBox.add(this.userLabel,
                    { x_fill:  true,
                        y_fill:  false,
                        x_align: St.Align.END,
                        y_align: St.Align.MIDDLE });
            this._user = AccountsService.UserManager.get_default().get_user(GLib.get_user_name());
            this._userLoadedId = this._user.connect('notify::is_loaded', Lang.bind(this, this._onUserChanged));
            this._userChangedId = this._user.connect('changed', Lang.bind(this, this._onUserChanged));
            this.actor = new St.Bin({style_class: 'user-icon'});
            this._icon = new St.Icon({
                icon_size: HOVER_ICON_SIZE,
                icon_type: St.IconType.FULLCOLOR
            });
            //this.actor.set_child(this._icon);
            this.actor.show();

            this.actor.set_child(userBox);

            this._onUserChanged();
        }
        catch(e) {
            global.logError(e);
        }

    },
  
    _onUserChanged: function() {
        this._icon.hide();
        if (this._user.is_loaded) {
            this.set_applet_tooltip(this._user.get_real_name());   
            this.userLabel.set_text (this._user.get_real_name());
            if (this._userIcon) {
                let iconFileName = this._user.get_icon_file();
                let iconFile = Gio.file_new_for_path(iconFileName);
                let icon;
                if (iconFile.query_exists(null)) {
                    icon = new Gio.FileIcon({file: iconFile});
                } else {
                    icon = new Gio.ThemedIcon({name: 'avatar-default'});
                }
                let img = St.TextureCache.get_default().load_gicon(null, icon, 48);
                this._userIcon.set_child (img);
                this.actor.set_child(this._userIcon);
                this._userIcon.show();               
            }
        }
    },
    _refresh: function (icon_name) {
        if(icon_name) {
            this._userIcon.hide();
            this.actor.set_child(this._icon);
            this._icon.set_icon_name(icon_name);
            this._icon.show();
        }
        else { 
            this._userIcon.show();
            this._icon.hide();
        }
    }
};
/*
function ShutdownContextMenuItem(parentMenu, menu, label, action) {
    this._init(parentMenu, menu, label, action);
}

ShutdownContextMenuItem.prototype = {
    __proto__: ApplicationContextMenuItem.prototype,

    _init: function (parentMenu, menu, label, action) {
        this.parentMenu = parentMenu;
        ApplicationContextMenuItem.prototype._init.call(this, menu, label, action);
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
    },

    activate: function (event) {
        switch (this._action) {
        case "logout":
            this._session.LogoutRemote(0);
            break;
        case "lock":
            this._screenSaverProxy.LockRemote();
            break;
        }
        this._appButton.toggle();
        this.parentMenu.toggle();
        return false;
    }

};
*/
function ShutdownMenu(parent, hoverIcon) {
    this._init(parent, hoverIcon);
}

ShutdownMenu.prototype = {
    __proto__: PopupMenu.PopupSubMenuMenuItem.prototype,

    _init: function (parent, hoverIcon) {
        this._session = new GnomeSession.SessionManager();
        this._screenSaverProxy = new ScreenSaver.ScreenSaverProxy();
        let label = '';
        this.hoverIcon = hoverIcon;
        this.parent = parent;
        PopupMenu.PopupSubMenuMenuItem.prototype._init.call(this, label);
        this.actor.set_style_class_name('menu-category-button');
        //this.removeActor(this.label);
        //this.removeActor(this._triangle);
        //this._triangle = new St.Label();
        let parentT = this.label.get_parent();
        if(parentT == this.actor)
            this.removeActor(this.label);
        if(parentT != null)
            parentT.remove_actor(this.label);
        parentT = this._triangle.get_parent();
        if(parentT == this.actor)
            this.removeActor(this._triangle);
        else if(parentT != null)
            parentT.remove_actor(this._triangle);

	    this.icon = new St.Icon({style_class: 'popup-menu-icon', icon_type: St.IconType.FULLCOLOR, icon_name: 'forward', icon_size: ICON_SIZE });
	    this.addActor(this.icon);

        this.menu = new ApplicationMenu(this.actor);
        this.menu.actor.remove_style_class_name("popup-sub-menu");
        /*
        let menuItem;
        menuItem = new ShutdownContextMenuItem(this.parent, this.menu, _("Logout"), "logout");
        this.menu.addMenuItem(menuItem);
        menuItem = new ShutdownContextMenuItem(this.parent, this.menu, _("Lock Screen"), "lock");
        this.menu.addMenuItem(menuItem);
        */
        let item;
        item = new PopupMenu.PopupIconMenuItem(_("Lock Screen"), "lock-screen", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            let screensaver_settings = new Gio.Settings({schema: "org.cdos.desktop.screensaver"});
            let screensaver_dialog   = Gio.file_new_for_path("/usr/bin/cdos-screensaver-command");
            if(screensaver_dialog.query_exists(null)) {
                if(screensaver_settings.get_boolean("ask-for-away-message")) {
                    Util.spawnCommandLine("cdos-screensaver-lock-dialog");
                }
                else {
                    Util.spawnCommandLine("cdos-screensaver-command --lock");
                }
            }
        }));
        this.menu.addMenuItem(item);
        if (GLib.getenv("XDG_SEAT_PATH")) {
            // LightDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-greeter");
            }));
            this.menu.addMenuItem(item);

            item = new PopupMenu.PopupIconMenuItem(_("Guest Session"), "guest-session", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("dm-tool switch-to-guest");
            }));
            this.menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/mdmflexiserver", GLib.FileTest.EXISTS)) {
            // MDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("mdmflexiserver");
            }));
            this.menu.addMenuItem(item);
        }
        else if (GLib.file_test("/usr/bin/gdmflexiserver", GLib.FileTest.EXISTS)) {
            // GDM
            item = new PopupMenu.PopupIconMenuItem(_("Switch User"), "switch-user", St.IconType.SYMBOLIC);
            item.connect('activate', Lang.bind(this, function() {
                Util.spawnCommandLine("cdos-screensaver-command --lock");
                Util.spawnCommandLine("gdmflexiserver");
                }));
                this.menu.addMenuItem(item);
        }
        item = new PopupMenu.PopupIconMenuItem(_("Log Out..."), "logout", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.LogoutRemote(0);
        }));
        this.menu.addMenuItem(item);

        this.menu.addMenuItem(new PopupMenu.PopupSeparatorMenuItem());

        item = new PopupMenu.PopupIconMenuItem(_("Power Off..."), "shutdown", St.IconType.SYMBOLIC);
        item.connect('activate', Lang.bind(this, function() {
            this._session.ShutdownRemote();
        }));
        this.menu.addMenuItem(item);
    },

    setActive: function (active) {
        if (active) { 
            this.actor.set_style_class_name('menu-category-button-selected');
            this.hoverIcon._refresh('system-log-out');
        }
        else this.actor.set_style_class_name('menu-category-button');
    },

    _onButtonReleaseEvent: function (actor, event) {
        if (event.get_button() == 1) {
            this.menu.toggle();
        }

    }
};



function RightButtonsBox(appsMenuButton, menu) {
    this._init(appsMenuButton, menu);
}

RightButtonsBox.prototype = {
    _init: function (appsMenuButton, menu) {
        this.appsMenuButton = appsMenuButton;
        this.actor = new St.BoxLayout({
            vertical: true
        });
        this.shutdownBox = new St.BoxLayout({
            vertical: false
        });
        this.actor._delegate = this;
        this.menu = menu;
        this.addItems();
    },

    addItems: function () {
        this.hoverIcon = new HoverIcon(this.menu);
        this.home = new TextBoxItem(_("Home"), "folder-home", "Util.spawnCommandLine('nemo')", this.menu, this.hoverIcon, false);
        this.documents = new TextBoxItem(_("Documents"), "folder-documents", "Util.spawnCommandLine('nemo Documents')", this.menu, this.hoverIcon, false);
        this.pictures = new TextBoxItem(_("Pictures"), "folder-pictures", "Util.spawnCommandLine('nemo Pictures')", this.menu, this.hoverIcon, false);
        this.music = new TextBoxItem(_("Music"), "folder-music", "Util.spawnCommandLine('nemo Music')", this.menu, this.hoverIcon, false);
        this.videos = new TextBoxItem(_("Videos"), "folder-videos", "Util.spawnCommandLine('nemo Videos')", this.menu, this.hoverIcon, false);
        this.computer = new TextBoxItem(_("Computer"), "computer", "Util.spawnCommandLine('nemo computer:///')", this.menu, this.hoverIcon, false);
        //this.packageItem = new Items.TextBoxItem(_("Package Manager"), "synaptic", "Util.spawnCommandLine('gksu synaptic')", this.menu, this.hoverIcon, false);
        this.control = new TextBoxItem(_("Control Center"), "gnome-control-center", "Util.spawnCommandLine('cdos-control-center')", this.menu, this.hoverIcon, false);
        this.run = new TextBoxItem(_("Run"), "system-run", "Main.getRunDialog().open()", this.menu, this.hoverIcon, false);
        this.terminal = new TextBoxItem(_("Terminal"), "terminal", "Util.spawnCommandLine('gnome-terminal')", this.menu, this.hoverIcon, false);
        this.help = new TextBoxItem(_("Help"), "help", "Util.spawnCommandLine('yelp')", this.menu, this.hoverIcon, false);
        this.shutdown = new TextBoxItem("Shutdown", "system-shutdown", "Session.ShutdownRemote()", this.menu, this.hoverIcon, false);
        this.shutdownMenu = new ShutdownMenu(this.menu, this.hoverIcon);

        this.shutdownBox.add_actor(this.shutdown.actor);
        this.shutdownBox.add_actor(this.shutdownMenu.actor);

        this.actor.add_actor(this.hoverIcon.actor);
        this.actor.add_actor(this.home.actor);
        this.actor.add_actor(this.documents.actor);
        this.actor.add_actor(this.pictures.actor);
        this.actor.add_actor(this.music.actor);
        this.actor.add_actor(this.videos.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.computer.actor);
        this.actor.add_actor(this.control.actor);
        //this.actor.add_actor(this.packageItem.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.run.actor);
        this.actor.add_actor(this.terminal.actor);
        this.actor.add_actor(this.help.actor);
        this.actor.add_actor(new PopupMenu.PopupSeparatorMenuItem().actor);
        this.actor.add_actor(this.shutdownBox);
        this.actor.add_actor(this.shutdownMenu.menu.actor);
    }
}

function FavoritesBox() {
    this._init();
}

FavoritesBox.prototype = {
    _init: function () {
        this.actor = new St.BoxLayout({
            vertical: true
        });
        this.actor._delegate = this;

        this._dragPlaceholder = null;
        this._dragPlaceholderPos = -1;
        this._animatingPlaceholdersCount = 0;
    },

    _clearDragPlaceholder: function () {
        if (this._dragPlaceholder) {
            this._dragPlaceholder.animateOutAndDestroy();
            this._dragPlaceholder = null;
            this._dragPlaceholderPos = -1;
        }
    },

    handleDragOver: function (source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed() || (!(source instanceof Items.FavoritesButton) && app.get_id() in AppFavorites.getAppFavorites().getFavoriteMap())) return DND.DragMotionResult.NO_DROP;

        let favorites = AppFavorites.getAppFavorites().getFavorites();
        let numFavorites = favorites.length;

        let favPos = favorites.indexOf(app);

        let children = this.actor.get_children();
        let numChildren = children.length;
        let boxHeight = this.actor.height;

        // Keep the placeholder out of the index calculation; assuming that
        // the remove target has the same size as "normal" items, we don't
        // need to do the same adjustment there.
        if (this._dragPlaceholder) {
            boxHeight -= this._dragPlaceholder.actor.height;
            numChildren--;
        }

        let pos = Math.round(y * numFavorites / boxHeight);

        if (pos != this._dragPlaceholderPos && pos <= numFavorites) {
            if (this._animatingPlaceholdersCount > 0) {
                let appChildren = children.filter(function (actor) {
                    return (actor._delegate instanceof Items.FavoritesButton);
                });
                this._dragPlaceholderPos = children.indexOf(appChildren[pos]);
            } else {
                this._dragPlaceholderPos = pos;
            }

            // Don't allow positioning before or after self
            if (favPos != -1 && (pos == favPos || pos == favPos + 1)) {
                if (this._dragPlaceholder) {
                    this._dragPlaceholder.animateOutAndDestroy();
                    this._animatingPlaceholdersCount++;
                    this._dragPlaceholder.actor.connect('destroy', Lang.bind(this, function () {
                        this._animatingPlaceholdersCount--;
                    }));
                }
                this._dragPlaceholder = null;

                return DND.DragMotionResult.CONTINUE;
            }

            // If the placeholder already exists, we just move
            // it, but if we are adding it, expand its size in
            // an animation
            let fadeIn;
            if (this._dragPlaceholder) {
                this._dragPlaceholder.actor.destroy();
                fadeIn = false;
            } else {
                fadeIn = true;
            }

            this._dragPlaceholder = new DND.GenericDragPlaceholderItem();
            this._dragPlaceholder.child.set_width(source.actor.height);
            this._dragPlaceholder.child.set_height(source.actor.height);
            this.actor.insert_actor(this._dragPlaceholder.actor, this._dragPlaceholderPos);
            if (fadeIn) this._dragPlaceholder.animateIn();
        }

        let srcIsFavorite = (favPos != -1);

        if (srcIsFavorite) return DND.DragMotionResult.MOVE_DROP;

        return DND.DragMotionResult.COPY_DROP;
    },

    // Draggable target interface
    acceptDrop: function (source, actor, x, y, time) {
        let app = source.app;

        // Don't allow favoriting of transient apps
        if (app == null || app.is_window_backed()) {
            return false;
        }

        let id = app.get_id();

        let favorites = AppFavorites.getAppFavorites().getFavoriteMap();

        let srcIsFavorite = (id in favorites);

        let favPos = 0;
        let children = this.actor.get_children();
        for (let i = 0; i < this._dragPlaceholderPos; i++) {
            if (this._dragPlaceholder && children[i] == this._dragPlaceholder.actor) continue;

            if (!(children[i]._delegate instanceof Items.FavoritesButton)) continue;

            let childId = children[i]._delegate.app.get_id();
            if (childId == id) continue;
            if (childId in favorites) favPos++;
        }

        Meta.later_add(Meta.LaterType.BEFORE_REDRAW, Lang.bind(this, function () {
            let appFavorites = AppFavorites.getAppFavorites();
            if (srcIsFavorite) appFavorites.moveFavoriteToPos(id, favPos);
            else appFavorites.addFavoriteAtPos(id, favPos);
            return false;
        }));

        return true;
    }
}
